/* This file includes ethernet task and its related functions. Please refer to 
 * the header file for description of functions. 
 */
 
/* Standard includes. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* Scheduler includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

/* Demo includes. */
#include "SAM7_EMAC.h"

/* lwIP includes. */
#include "lwip/api.h" 
#include "lwip/tcpip.h"
#include "lwip/memp.h" 
#include "lwip/stats.h"
#include "netif/loopif.h"
#include "pio.h"
#include "mac_tasks.h"
#include "adc.h"
#include "spi.h"
#include "ADS1258.h"
#include "EMflash.h"

static struct tcp_pcb *netio_pcb;
/* comm_open is used to hold the state of connection. When we are connected it
 * becomes 1. Then it is used to control the state of connection. If it is less
 * than 1 we don't send any data from ethernet. 
 */
static unsigned char comm_open;

/* config_mode is used to control the configuration mode of the device
 * from ethernet connection.
 */
static char config_mode = 0;

void adc_monitor_via_TCP( void *pvParameters )
{ 	
	/* MAC related variables */
	struct ip_addr my_ip_addr, my_net_mask, my_gateway;
	extern err_t ethernetif_init( struct netif *netif );
	static struct netif emac_if;
			
	/* Parameters are not used - suppress compiler error. */
	( void ) pvParameters;	
 			  		
	/* Create a binary semaphore to use common data buffer with spi 
	 * This semaphore is taken from ISR when the buffer is filled 
	 */
	vSemaphoreCreateBinary( spi_buffer_semaphore );
	if( spi_buffer_semaphore == NULL ) { 
		while( true )
			pioA_set_led( 0 , 0 ); 
	} 	
	/* No communication avaliable in the beginning */
	comm_open = 0 ;
		
	/* Create and configure the EMAC interface. */
	
	IP4_ADDR(&my_ip_addr,(SETTINGS_PAGE[LOCAL_IP] & 0xFF000000) >> 24 \
						,(SETTINGS_PAGE[LOCAL_IP] & 0x00FF0000) >> 16 \
						,(SETTINGS_PAGE[LOCAL_IP] & 0x0000FF00) >> 8 \
						,(SETTINGS_PAGE[LOCAL_IP] & 0x000000FF) >> 0);
	
	IP4_ADDR(&my_net_mask,(SETTINGS_PAGE[NETMASK] & 0xFF000000) >> 24 \
						 ,(SETTINGS_PAGE[NETMASK] & 0x00FF0000) >> 16 \
						 ,(SETTINGS_PAGE[NETMASK] & 0x0000FF00) >> 8 \
						 ,(SETTINGS_PAGE[NETMASK] & 0x000000FF) >> 0);	
	IP4_ADDR(&my_gateway,emacGATEWAY_ADDR0,emacGATEWAY_ADDR1,emacGATEWAY_ADDR2,emacGATEWAY_ADDR3);	
	netif_add(&emac_if, &my_ip_addr, &my_net_mask, &my_gateway, NULL, ethernetif_init, tcpip_input);
			
	/* make it the default interface */
    netif_set_default(&emac_if);

	/* bring it up */
    netif_set_up(&emac_if);
    
    /* We are using then raw api of LWiP for tcp connections. You can refer to 
     * netio_ functions at the end. All listening,accepting,receiving events
     * handled using callback functions. We are initializing the sequence using
     * netio_init then stack directs us when a event occurs
     */	 
	netio_init();
	/* No sending until we receive a "send" command */    
    comm_open = 1;
	for( ;; )
	{
		if( xSemaphoreTake( spi_buffer_semaphore, ( portTickType ) 0xFFFF ) == pdTRUE ) {
			if( comm_open > 1 ) {
				/* The trick for increasing TCP performance is calling tcp_output
				 * function after tcp_write provided that windows,send buffers are
				 * big enough.
				 */				
				if(tcp_write(netio_pcb, (int *)filter(), DATA_FRAGMENT_LENGTH, 0) == ERR_OK) {
					tcp_output(netio_pcb);					
				}								
			}							
		} 
	}
}
/* Please note that we won't send ack to each message since TCP stack merges small
 * packets into one. So we will only send compulsory messages
 */ 
inline void process_tcp_message( int *settings , struct tcp_pcb *pcb, struct pbuf *p )
{
	char *mes = (char *)(p->payload);
	int i;
	if( strncasecmp( mes , "Echo" , 4) == 0 ) {
		send_tcp_message("Sure!\0" , pcb);												 				 				    									
	}
	/* Enter configuration mode if requested */
	if( strncasecmp( mes , "Enter Config" , 12) == 0 ) {
		ADS1258_start_stop(0);
		/* When we are in config mode config_mode is 1 */
		config_mode = 1;				
		//send_tcp_message("Entering configuration mode...\0" , pcb);
		if(settings[0] != 0x10) {
			/* If this is the first time we enter config mode than initialize
			 * settings. Otherwise do not change it so that user can test settings
			 * before flashing settings.
			 */			 		
			for( i = 0; i < SETTINGS_LENGTH; i++ )
				settings[i] = SETTINGS_PAGE[i];
		}
		//send_tcp_block( (int *)settings , SETTINGS_LENGTH , pcb ); 				 				    									
	}
	/* Start sending ADC data */
	if( strncasecmp( mes , "Send" , 4) == 0 ) {
		comm_open = 2;				 				 				    									
	}
	/* Stop sending ADC data */
	if( strncasecmp( mes , "Stop sending" , 12) == 0 ) {
		comm_open = 1;				 				 				    									
	}							
	/* This function returns avaliable RAM in the FREERTOS heap */	  
	if( strncasecmp( mes , "Mem Test" , 8) == 0 ) {
		mes[0] = (char)vPortUsedMem() ;
		mes[1] = (char)(vPortUsedMem() >> 8) ;				
		tcp_write(pcb, p->payload,sizeof(p->payload), 1);
		tcp_output(pcb);								
	}
}
inline void configure_device_tcp( int *settings , struct tcp_pcb *pcb, struct pbuf *p )
{
	int i;
	char *mes = (char *)(p->payload);
					
	if( strncasecmp( mes , "Update firmware" , 15) == 0 ) {
		mes[0] = boot_from_rom() + 48 ;
		//tcp_write(pcb, p->payload, sizeof(p->payload), 1);
		//tcp_output(pcb);										 				 				    									
	}			
	/* Write settings to the flash. */
	if( strncasecmp( mes , "Save config" , 11) == 0 ) {
		mes[0] = flash_write( settings , SETTINGS_LENGTH , SETTINGS_PAGE_NUMBER ) + 48 ;			
		/* Initialize adc sub-system to reflect changes */
		adc_system_init();	
		//tcp_write(pcb, p->payload, sizeof(p->payload), 1);
		//tcp_output(pcb);		
	}
	/* Send settings to the PC */
	if( strncasecmp( mes , "Read all settings" , 17) == 0 ) {
		send_tcp_block( (int *)settings , SETTINGS_LENGTH , pcb );					
	}
	/* Receive settings from PC */
	if( strncasecmp( mes , "Write all settings" , 18) == 0 ) {
		/* Message length should be settings length in bytes plus 20 bytes */		
		if ( p->len >= 20 +  SETTINGS_LENGTH * 4 ) {				
			for( i = 0; i < SETTINGS_LENGTH ; i++ )
				settings[i] = ( mes[20 + i * 4] << 0 )  \
							+ ( mes[21 + i * 4] << 8 )  \
							+ ( mes[22 + i * 4] << 16 ) \
							+ ( mes[23 + i * 4] << 24 );		
			//mes[0] = flash_write( settings , SETTINGS_LENGTH , SETTINGS_PAGE_NUMBER ) + 48 ;			
			/* Initialize adc sub-system to reflect changes */
			//adc_system_init();
			//ADS1258_start_stop(1);
			//config_mode = 0;											
			//send_tcp_message("All settings changed\0" , pcb); 
		} else {
			//send_tcp_message("All settings change fails!\0" , pcb);
		}					
	}
	/* This command corrects offsets of the ADC sub-system */
	if( strncasecmp( mes , "Correct Offsets" , 15) == 0 ) {
		if( offset_correct( &settings[0] ) == 0 ) 
			send_tcp_message("Offsets are corrected\0" , pcb);
		else
			send_tcp_message("Offset corrections failed\0" , pcb);						
	}
	/* Exit config and start ADC sub-system */
	if( strncasecmp( mes , "Exit Config" , 11) == 0 ) {
		ADS1258_start_stop(1);
		config_mode = 0;			
		//send_tcp_message("Exiting configuration mode...\0" , pcb);						
	}
}
inline void send_tcp_message(char *mes , struct tcp_pcb *pcb) 
{	
	/* We use abstraction layer for this functions for future use. There may somethings
	 * before and after we call tcp_write. Just see that this function is inlined for 
	 * now. 
	 */
	tcp_write(pcb, mes, strlen(mes), 1);
	tcp_output(pcb);
}
inline void send_tcp_block(int *block , short length , struct tcp_pcb *pcb) 
{
	/* We use abstraction layer for this functions for future use. There may somethings
	 * before and after we call tcp_write. Just see that this function is inlined for 
	 * now. 
	 */
	tcp_write(pcb, block, sizeof(int) * length, 1);
	tcp_output(pcb);	
	
}
err_t netio_sent(void *arg, struct tcp_pcb *tpcb, u16_t len)
{
	(void) arg;
	(void) tpcb;
	(void) len;	
	
	return ERR_OK;
}
err_t netio_recv(void *arg, struct tcp_pcb *pcb, struct pbuf *p, err_t err)
{
	/* This callback function is called when stack receives data from the MAC
	 * interface(on our port of course). Then we can process this message
	 */  
	static int settings[SETTINGS_LENGTH];
	(void) arg;
	if (err == ERR_OK && p != NULL)
	{
		/* if config_mode is not 0, then we are in configuration mode */
		if( config_mode ) {
			configure_device_tcp( &settings[0] , pcb , p );		
		} else {
			/* Process message then if we are not in config mode */
			process_tcp_message( &settings[0] , pcb , p );													 				 				    									
		}
		/* Acknowledge stack that we have processed the message so that it can 
		 * inrcrease window size 
		,
		 */
		tcp_recved(pcb, p->tot_len);
		/* Never forget to free a buffer! */
		pbuf_free(p);		
	}
	else
		pbuf_free(p); /* Never forget to free a buffer! */

	if (err == ERR_OK && p == NULL)
	{
		/* If we are here this means that other side has closed the connection. */
		tcp_arg(pcb, NULL);
		tcp_sent(pcb, NULL);
		tcp_recv(pcb, NULL);
		tcp_close(pcb);		
		/* Inform mac task that communication is closed. Otherwise, it will immediately
		 * send conversion data as soon as a client arrives without waiting the
		 * send commdand.
		 */
		comm_open = 0;
	}
	return ERR_OK;
}

err_t netio_accept(void *arg, struct tcp_pcb *pcb, err_t err)
{
	/* When there is a connection request this function is called by the stack */
	(void) arg;
	(void) err;
	tcp_arg(pcb, NULL);
	tcp_sent(pcb, netio_sent);
	/* Inform stack about which function to call-back when there is data on the line */ 
	tcp_recv(pcb, netio_recv);
	/* Hold the accepted connection pcb. This will used later when we desire to send
	 * data asynchronously from the receive callback function
	 */
	netio_pcb = pcb;	
	return ERR_OK;
}

struct tcp_pcb* netio_init(void)
{
	/* This function initalizes the connection. We start listenin on the desired port
	 * and our defined callback function to be called by the stack when someone 
	 * arrives
	 */ 
	struct tcp_pcb *pcb;
	
	pcb = tcp_new();
	tcp_bind(pcb, IP_ADDR_ANY, SETTINGS_PAGE[LOCAL_PORT]);
	pcb = tcp_listen(pcb);	
	tcp_accept(pcb, netio_accept);
	return pcb;
}
